//-----------------------------------------------------------------------
// <copyright file="MockObject.cs" company="NMock2">
//
//   http://www.sourceforge.net/projects/NMock2
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
// </copyright>
//-----------------------------------------------------------------------
namespace NMock2.Internal
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Reflection;
    using Monitoring;

    public class MockObject : IInvokable, IMockObject
    {
        /// <summary>
        /// Stores the backlink to the mockery which created this mock object.
        /// </summary>
        private readonly Mockery mockery;

        /// <summary>
        /// Stores the mocked type.
        /// </summary>
        private readonly Type mockedType;

        /// <summary>
        /// Stores the name of the mock object.
        /// </summary>
        private readonly string name;

        /// <summary>
        /// Stores the event handlers that could be added to the mock object.
        /// </summary>
        private readonly IDictionary eventHandlers;

		public string MockName
		{
			get { return name; }
		}

		protected Type MockedType
		{
			get { return mockedType; }
		}

		protected Mockery Mockery
		{
			get { return mockery; }
		}

        protected MockObject(Mockery mockery, Type mockedType, string name)
        {
            this.mockery = mockery;
            this.mockedType = mockedType;
            this.name = name;
            this.eventHandlers = new Hashtable();
        }

        public override string ToString()
        {
            return this.name;
        }

        public void Invoke(Invocation invocation)
        {
            this.mockery.Dispatch(invocation);
        }

        public bool HasMethodMatching(Matcher methodMatcher)
        {
			return Mockery.GetMatchingMethodsForType(mockedType, methodMatcher, true).Count > 0;
        }

		public IList<MethodInfo> GetMethodsMatching(Matcher methodMatcher)
		{
			return Mockery.GetMatchingMethodsForType(mockedType, methodMatcher, false);
		}

        public void AddExpectation(IExpectation expectation)
        {
            this.mockery.AddExpectation(expectation);
        }

        public void AddEventHandler(string eventName, Delegate handler)
        {
            ArrayList handlers = (ArrayList)this.eventHandlers[eventName];

            if (handlers == null)
            {
                handlers = new ArrayList();
                this.eventHandlers.Add(eventName, handlers);
            }

            if (! handlers.Contains(handler)) 
            { 
                handlers.Add(handler); 
            }
        }

        public void RaiseEvent(string eventName, params object[] args)
        {
            IEnumerable handlers = (IEnumerable)this.eventHandlers[eventName];
    
            if (handlers != null)
            {
                // copy handlers before invocation to prevent colection modified exception if event handler is removed within event handler itself.
                List<Delegate> delegates = new List<Delegate>();
                foreach (Delegate handler in handlers)
                {
                    delegates.Add(handler);    
                }

                foreach (Delegate d in delegates)
                {
                    try
                    {
                        d.DynamicInvoke(args);
                    }
                    catch (TargetInvocationException tie)
                    {
                        // replace stack trace with original stack trace
                        FieldInfo remoteStackTraceString = typeof(Exception).GetField("_remoteStackTraceString", BindingFlags.Instance | BindingFlags.NonPublic);
                        remoteStackTraceString.SetValue(tie.InnerException, tie.InnerException.StackTrace + Environment.NewLine);
                        throw tie.InnerException;
                    }
                }
            }
            else
            {
                // check whether the mocked type has this event
                if (this.mockedType.GetEvent(eventName) == null)
                {
                    // check whether an inherited interface has this event
                    bool found = false;
                    foreach (Type type in this.mockedType.GetInterfaces())
                    {
                        if (type.GetEvent(eventName) != null)
                        {
                            found = true;
                            break;
                        }
                    }

                    if (!found)
                    {
                        throw new ArgumentException(
                            string.Format("Event with name {0} does not exist on type {1}.", eventName,
                                          this.mockedType.FullName),
                            "eventName");
                    }
                }
            }
        }

        public void RemoveEventHandler(string eventName, Delegate handler)
        {
            ArrayList handlers = (ArrayList)this.eventHandlers[eventName];

            if (handlers != null)
            {
                handlers.Remove(handler);
                if (handlers.Count == 0)
                {
                    this.eventHandlers.Remove(eventName);
                }
            }
        }
    }
}
