//-----------------------------------------------------------------------
// <copyright file="Mockery.cs" company="NMock2">
//
//   http://www.sourceforge.net/projects/NMock2
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
// </copyright>
//-----------------------------------------------------------------------

namespace NMock2
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Reflection;
    using Internal;
    using Monitoring;

    /// <summary>
    /// The mockery is used to create dynamic mocks and check that all expectations were met during a unit test.
    /// </summary>
    /// <remarks>Name inspired by Ivan Moore.</remarks>
    public class Mockery : IDisposable
    {
        /// <summary>
        /// In the rare case where the default mock object factory is replaced, we hold on to the
        /// previous factory (or factories) in case we need to switch back to them.
        /// </summary>
        private static readonly Dictionary<Type, IMockObjectFactory> availableMockObjectFactories = new Dictionary<Type, IMockObjectFactory>();

        /// <summary>
        /// The mock object factory that is being used by this Mockery instance.
        /// </summary>
        private readonly IMockObjectFactory currentMockObjectFactory;

        /// <summary>
        /// The mock object factory that will be used when a new Mockery instance is created
        /// </summary>
        private static IMockObjectFactory defaultMockObjectFactory;

        /// <summary>
        /// Depth of cascaded ordered, unordered expecation blocks.
        /// </summary>
        private int depth;

        /// <summary>
        /// All expectations.
        /// </summary>
        private IExpectationOrdering expectations;

        /// <summary>
        /// Expectations at current cascade level.
        /// </summary>
        private IExpectationOrdering topOrdering;

        /// <summary>
        /// If an unexpected invocation exception is thrown then it is stored here to rethrow it in the 
        /// <see cref="VerifyAllExpectationsHaveBeenMet"/> method - exception cannot be swallowed by tested code.
        /// </summary>
        private ExpectationException thrownUnexpectedInvocationException;

        /// <summary>
        /// Initializes static members of the <see cref="Mockery"/> class.
        /// </summary>
        static Mockery()
        {
            ChangeDefaultMockObjectFactory(typeof(CastleMockObjectFactory));
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Mockery"/> class.
        /// Clears all expectations.
        /// </summary>
        public Mockery()
        {
            this.currentMockObjectFactory = defaultMockObjectFactory;
            
            this.ClearExpectations();
        }

        /// <summary>
        /// Gets a disposabale object and tells the mockery that the following expectations are ordered, i.e. they have to be met in the specified order.
        /// Dispose the returned value to return to previous mode.
        /// </summary>
        /// <value>Disposable object. When this object is disposed then the ordered expectation mode is set back to the mode it was previously
        /// to call to <see cref="Ordered"/>.</value>
        public IDisposable Ordered
        {
            get { return this.Push(new OrderedExpectations(this.depth)); }
        }

        /// <summary>
        /// Gets a disposabale object and tells the mockery that the following expectations are unordered, i.e. they can be met in any order.
        /// Dispose the returned value to return to previous mode.
        /// </summary>
        /// <value>Disposable object. When this object is disposed then the unordered expectation mode is set back to the mode it was previously
        /// to the call to <see cref="Unordered"/>.</value>
        public IDisposable Unordered
        {
            get { return this.Push(new UnorderedExpectations(this.depth)); }
        }

        /// <summary>
        /// Allows the default IMockObjectFactory to be replaced with a different implementation.
        /// </summary>
        /// <param name="factoryType">The System.Type of the IMockObjectFactory implementation to use.
        /// This is expected to implement IMockObjectFactory and have a default constructor.</param>
        public static void ChangeDefaultMockObjectFactory(Type factoryType)
        {
            if (!typeof(IMockObjectFactory).IsAssignableFrom(factoryType))
            {
                throw new ArgumentException("Supplied factory type does not implement IMockObjectFactory", "factoryType");
            }

            lock (availableMockObjectFactories)
            {
                if (!availableMockObjectFactories.TryGetValue(factoryType, out defaultMockObjectFactory))
                {
                    try
                    {
                        defaultMockObjectFactory = (IMockObjectFactory)Activator.CreateInstance(factoryType);
                        availableMockObjectFactories[factoryType] = defaultMockObjectFactory;
                    }
                    catch (MissingMethodException)
                    {
                        throw new ArgumentException("Supplied factory type does not have a default constructor", "factoryType");
                    }
                }
            }
        }

        /// <summary>
        /// Creates a new dynamic mock of the specified type.
        /// </summary>
        /// <param name="mockedType">The type to mock.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A dynamic mock for the specified type.</returns>
        public object NewMock(Type mockedType, params object[] constructorArgs)
        {
            return this.CreateMockForType(mockedType, null, MockStyle.Default, constructorArgs);
        }

        /// <summary>
        /// Creates a new dynamic mock of the specified type.
        /// </summary>
        /// <param name="mockedType">The type to mock.</param>
        /// <param name="mockStyle">Specifies how the mock object should behave when first created.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A named dynamic mock for the specified type.</returns>
        public object NewMock(Type mockedType, MockStyle mockStyle, params object[] constructorArgs)
        {
            return this.CreateMockForType(mockedType, null, mockStyle, constructorArgs);
        }

        /// <summary>
        /// Creates a new dynamic mock of the specified type.
        /// </summary>
        /// <typeparam name="MockedType">The type to mock.</typeparam>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A dynamic mock for the specified type.</returns>
        public MockedType NewMock<MockedType>(params object[] constructorArgs)
        {
            return (MockedType)this.CreateMockForType(typeof(MockedType), null, MockStyle.Default, constructorArgs);
        }

        /// <summary>
        /// Creates a new dynamic mock of the specified type.
        /// </summary>
        /// <typeparam name="MockedType">The type to mock.</typeparam>
        /// <param name="mockStyle">Specifies how the mock object should behave when first created.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A dynamic mock for the specified type.</returns>
        public MockedType NewMock<MockedType>(MockStyle mockStyle, params object[] constructorArgs)
        {
            return (MockedType)this.CreateMockForType(typeof(MockedType), null, mockStyle, constructorArgs);
        }
        
        /// <summary>
        /// Creates a new named dynamic mock of the specified type.
        /// </summary>
        /// <param name="mockedType">The type to mock.</param>
        /// <param name="name">A name for the mock that will be used in error messages.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A named mock.</returns>
        public object NewNamedMock(Type mockedType, string name, params object[] constructorArgs)
        {
            return this.CreateMockForType(mockedType, name, MockStyle.Default, constructorArgs);
        }

        /// <summary>
        /// Creates a new named dynamic mock of the specified type and allows the style
        /// of the mock to be specified.
        /// </summary>
        /// <param name="mockedType">The type to mock.</param>
        /// <param name="name">A name for the mock that will be used in error messages.</param>
        /// <param name="mockStyle">Specifies how the mock object should behave when first created.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A named mock.</returns>
        public object NewNamedMock(Type mockedType, string name, MockStyle mockStyle, params object[] constructorArgs)
        {
            return this.CreateMockForType(mockedType, name, mockStyle, constructorArgs);
        }

        /// <summary>
        /// Creates a new named dynamic mock of the specified type.
        /// </summary>
        /// <typeparam name="MockedType">The type to mock.</typeparam>
        /// <param name="name">A name for the mock that will be used in error messages.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A named mock.</returns>
        public MockedType NewNamedMock<MockedType>(string name, params object[] constructorArgs)
        {
            return (MockedType)this.CreateMockForType(typeof(MockedType), name, MockStyle.Default, constructorArgs);
        }

        /// <summary>
        /// Creates a new named dynamic mock of the specified type and allows the style
        /// of the mock to be specified.
        /// </summary>
        /// <typeparam name="MockedType">The type to mock.</typeparam>
        /// <param name="name">A name for the mock that will be used in error messages.</param>
        /// <param name="mockStyle">Specifies how the mock object should behave when first created.</param>
        /// <param name="constructorArgs">The arguments for the constructor of the class to be mocked.
        /// Only applicable when mocking classes with non-default constructors.</param>
        /// <returns>A named mock.</returns>
        public MockedType NewNamedMock<MockedType>(string name, MockStyle mockStyle, params object[] constructorArgs)
        {
            return (MockedType)this.CreateMockForType(typeof(MockedType), name, mockStyle, constructorArgs);
        }

        /// <summary>
        /// Verifies that all expectations have been met.
        /// Will be called in <see cref="Dispose"/>, too. 
        /// </summary>
        public void VerifyAllExpectationsHaveBeenMet()
        {
            // check for swallowed exception
            if (this.thrownUnexpectedInvocationException != null)
            {
                Exception exceptionToBeRethrown = this.thrownUnexpectedInvocationException;
                this.thrownUnexpectedInvocationException = null; // only rethrow once

                throw new ExpectationException(
                    string.Format(
                        "Rethrown unexpected invocation exception. Stack trace of inner exception:{0}{1}",
                        Environment.NewLine,
                        exceptionToBeRethrown.StackTrace),
                    exceptionToBeRethrown);
            }

            if (!this.expectations.HasBeenMet)
            {
                this.FailUnmetExpectations();
            }
        }

        /// <summary>
        /// Disposes the mockery be verifying that all expectations were met.
        /// </summary>
        public void Dispose()
        {
            this.VerifyAllExpectationsHaveBeenMet();
        }

        /// <summary>
        /// Gets the methods of the specified type that match the specified matcher.
        /// </summary>
        /// <param name="type">The type to inspect.</param>
        /// <param name="matcher">The matcher.</param>
        /// <param name="firstMatchOnly">if set to <c>true</c> then only the first match is returned.</param>
        /// <returns>The methods of the specified type that match the specified matcher.</returns>
        internal static IList<MethodInfo> GetMatchingMethodsForType(Type type, Matcher matcher, bool firstMatchOnly)
        {
            List<MethodInfo> matches = new List<MethodInfo>();

            if (type.IsInterface)
            {
                foreach (Type implementedInterface in GetInterfacesImplementedByType(type))
                {
                    foreach (MethodInfo method in implementedInterface.GetMethods())
                    {
                        if (matcher.Matches(method))
                        {
                            matches.Add(method);

                            if (firstMatchOnly)
                            {
                                return matches;
                            }
                        }
                    }
                }
            }
            else
            {
                foreach (MethodInfo method in type.GetMethods(BindingFlags.Instance | BindingFlags.NonPublic | BindingFlags.Public))
                {
                    if (matcher.Matches(method))
                    {
                        matches.Add(method);

                        if (firstMatchOnly)
                        {
                            return matches;
                        }
                    }
                }
            }

            return matches;
        }

        /// <summary>
        /// Adds the expectation.
        /// </summary>
        /// <param name="expectation">The expectation.</param>
        internal void AddExpectation(IExpectation expectation)
        {
            this.topOrdering.AddExpectation(expectation);
        }

        /// <summary>
        /// Dispatches the specified invocation.
        /// </summary>
        /// <param name="invocation">The invocation.</param>
        internal void Dispatch(Invocation invocation)
        {
            if (this.expectations.Matches(invocation))
            {
                this.expectations.Perform(invocation);
            }
            else
            {
                this.FailUnexpectedInvocation(invocation);
            }
        }

        /// <summary>
        /// Determines whether there exist expectations for the specified invocation.
        /// </summary>
        /// <param name="invocation">The invocation.</param>
        /// <returns><c>true</c> if there exist expectations for the specified invocation; otherwise, <c>false</c>.
        /// </returns>
        internal bool HasExpectationFor(Invocation invocation)
        {
            return this.expectations.Matches(invocation);
        }

        /// <summary>
        /// Returns the default name for a type that is used to name mocks.
        /// </summary>
        /// <param name="type">The type to get the default name for.</param>
        /// <returns>Default name for the specified type.</returns>
        protected virtual string DefaultNameFor(Type type)
        {
            string name = type.Name;
            int firstLower = FirstLowerCaseChar(name);

            return 
                firstLower == name.Length ? 
                name.ToLower() : 
                name.Substring(firstLower - 1, 1).ToLower() + name.Substring(firstLower);
        }

        /// <summary>
        /// Gets the interfaces implemented by the specified type.
        /// </summary>
        /// <param name="type">The type to inspect.</param>
        /// <returns>The interfaces implemented by the specified type.</returns>
        private static Type[] GetInterfacesImplementedByType(Type type)
        {
            ArrayList implementedTypes = new ArrayList();
            foreach (Type implementedInterface in type.GetInterfaces())
            {
                implementedTypes.AddRange(GetInterfacesImplementedByType(implementedInterface));
            }

            implementedTypes.Add(type);

            Type[] types = new Type[implementedTypes.Count];
            implementedTypes.CopyTo(types);

            return types;
        }

        /// <summary>
        /// Finds the first lower case char in the specified string.
        /// </summary>
        /// <param name="s">The string to inspect.</param>
        /// <returns>the first lower case char in the specified string.</returns>
        private static int FirstLowerCaseChar(string s)
        {
            int i = 0;
            while (i < s.Length && !Char.IsLower(s[i]))
            {
                i++;
            }

            return i;
        }

        /// <summary>
        /// Checks that the interface does not contain ToString method declaration.
        /// </summary>
        /// <param name="mockedType">Type of the mocked.</param>
        private static void CheckInterfaceDoesNotContainToStringMethodDeclaration(Type mockedType)
        {
            MethodInfo methodInfo = mockedType.GetMethod("ToString");
            if (methodInfo != null && methodInfo.GetParameters().Length == 0)
            {
                throw new ArgumentException("Interfaces must not contain a declaration for ToString().");
            }
        }

        /// <summary>
        /// Clears the expectations.
        /// </summary>
        private void ClearExpectations()
        {
            this.depth = 1;
            this.expectations = new UnorderedExpectations();
            this.topOrdering = this.expectations;
        }

        /// <summary>
        /// Creates a new mock for the specified type.
        /// </summary>
        /// <param name="mockedType">Type of the mock.</param>
        /// <param name="name">The name of the mock.</param>
        /// <param name="mockStyle">The mock style.</param>
        /// <param name="constructorArgs">The constructor arguments.</param>
        /// <returns>A newly created mock for type</returns>
        private object CreateMockForType(Type mockedType, string name, MockStyle mockStyle, params object[] constructorArgs)
        {
            if (name == null)
            {
                name = this.DefaultNameFor(mockedType);
            }
            
            if (mockedType.IsInterface)
            {
                if (constructorArgs.Length > 0)
                {
                    throw new InvalidOperationException("Cannot specify constructor arguments when mocking an interface");
                }

                CheckInterfaceDoesNotContainToStringMethodDeclaration(mockedType);

                return this.currentMockObjectFactory.CreateMockForInterface(this, mockedType, name);
            }

            if (mockedType.IsClass)
            {
                return this.currentMockObjectFactory.CreateMockForClass(this, mockedType, name, mockStyle, constructorArgs);
            }

            throw new ArgumentException("Can only mock classes and interfaces", "mockedType"); // TODO: More error checking
        }

        /// <summary>
        /// Pushes the specified new ordering on the expectations stack.
        /// </summary>
        /// <param name="newOrdering">The new ordering.</param>
        /// <returns>Disposable popper.</returns>
        private Popper Push(IExpectationOrdering newOrdering)
        {
            this.topOrdering.AddExpectation(newOrdering);
            IExpectationOrdering oldOrdering = this.topOrdering;
            this.topOrdering = newOrdering;
            this.depth++;
            return new Popper(this, oldOrdering);
        }

        /// <summary>
        /// Pops the specified old ordering from the expectations stack.
        /// </summary>
        /// <param name="oldOrdering">The old ordering.</param>
        private void Pop(IExpectationOrdering oldOrdering)
        {
            this.topOrdering = oldOrdering;
            this.depth--;
        }

        /// <summary>
        /// Throws an exception listing all unmet expectations.
        /// </summary>
        private void FailUnmetExpectations()
        {
            DescriptionWriter writer = new DescriptionWriter();
            writer.WriteLine("not all expected invocations were performed");
            this.expectations.DescribeUnmetExpectationsTo(writer);
            this.ClearExpectations();

            throw new ExpectationException(writer.ToString());
        }

        /// <summary>
        /// Throws an exception indicating that the specified invocation is not expected.
        /// </summary>
        /// <param name="invocation">The invocation.</param>
        private void FailUnexpectedInvocation(Invocation invocation)
        {
            DescriptionWriter writer = new DescriptionWriter();
            writer.Write("unexpected invocation of ");
            invocation.DescribeTo(writer);
            writer.WriteLine();
            this.expectations.DescribeActiveExpectationsTo(writer);

            // try catch to get exception with stack trace.
            try
            {
                throw new ExpectationException(writer.ToString());
            }
            catch (ExpectationException e)
            {
                // remember only first exception
                if (this.thrownUnexpectedInvocationException == null)
                {
                    this.thrownUnexpectedInvocationException = e;
                }

                throw;
            }
        }

        /// <summary>
        /// A popper pops an expectation ordering from the expectations stack on disposal.
        /// </summary>
        private class Popper : IDisposable
        {
            /// <summary>
            /// The mockery.
            /// </summary>
            private readonly Mockery mockery;

            /// <summary>
            /// The previous expectaion ordering.
            /// </summary>
            private readonly IExpectationOrdering previous;

            /// <summary>
            /// Initializes a new instance of the <see cref="Popper"/> class.
            /// </summary>
            /// <param name="mockery">The mockery.</param>
            /// <param name="previous">The previous.</param>
            public Popper(Mockery mockery, IExpectationOrdering previous)
            {
                this.previous = previous;
                this.mockery = mockery;
            }

            /// <summary>
            /// Pops the expectation ordering from the stack.
            /// </summary>
            public void Dispose()
            {
                this.mockery.Pop(this.previous);
            }
        }
    }
}
