using System;
using System.Collections.Generic;
using System.Text;

using Castle.Core.Interceptor;
using Castle.DynamicProxy;
using NMock2.Internal;

namespace NMock2.Monitoring
{
	public class CastleMockObjectFactory : IMockObjectFactory
	{
		private Dictionary<Type, Type> cachedProxyTypes = new Dictionary<Type, Type>();


		#region IMockObjectFactory Members

		public object CreateMockForInterface(Mockery mockery, Type mockedType, string name)
		{
			if(!mockedType.IsInterface)
				throw new ArgumentException("Expected an interface Type.", "mockedType");

			IInterceptor interceptor = new MockObjectInterceptor(mockery, mockedType, name, MockStyle.Default);
			object[] activationArgs = new object[] { new IInterceptor[] { interceptor }, new object(), name };

			return Activator.CreateInstance(GetProxyTypeForInterface(mockedType), activationArgs);
		}

		public object CreateMockForClass(Mockery mockery, Type mockedType, string name, MockStyle mockStyle, object[] constructorArgs)
		{
			if(!mockedType.IsClass)
				throw new ArgumentException("Expected a class Type.", "mockedType");

			if(mockedType.IsSealed)
				throw new ArgumentException("Cannot mock sealed classes.", "mockedType");

			IInterceptor interceptor = new MockObjectInterceptor(mockery, mockedType, name, mockStyle);
			object[] activationArgs = new object[constructorArgs.Length + 1];

			constructorArgs.CopyTo(activationArgs, 1);
			activationArgs[0] = new IInterceptor[] { interceptor };

			return Activator.CreateInstance(GetProxyTypeForClass(mockedType), activationArgs);
		}

		#endregion

		private Type GetProxyTypeForInterface(Type interfaceType)
		{
			if(!cachedProxyTypes.ContainsKey(interfaceType))
			{
				DefaultProxyBuilder proxyBuilder = new DefaultProxyBuilder();

				ProxyGenerationOptions options = new ProxyGenerationOptions();
				options.BaseTypeForInterfaceProxy = typeof(InterfaceMockBase);

				Type proxyType = proxyBuilder.CreateInterfaceProxyTypeWithoutTarget(interfaceType,
																					new Type[] { typeof(IMockObject) },
																					options);

				cachedProxyTypes[interfaceType] = proxyType;
			}

			return cachedProxyTypes[interfaceType];
		}

		private Type GetProxyTypeForClass(Type classType)
		{
			if(!cachedProxyTypes.ContainsKey(classType))
			{
				DefaultProxyBuilder proxyBuilder = new DefaultProxyBuilder();
				Type proxyType = proxyBuilder.CreateClassProxy(	classType,
																new Type[] { typeof(IMockObject) },
																ProxyGenerationOptions.Default);

				cachedProxyTypes[classType] = proxyType;
			}

			return cachedProxyTypes[classType];
		}




		/// <summary>
		/// Used as a base for interface mocks in order to provide a holder
		/// for a meaningful ToString() value.
		/// </summary>
		public class InterfaceMockBase
		{
			private string stringValue;

			public InterfaceMockBase(string stringValue)
			{
				this.stringValue = stringValue;
			}

			public override string ToString()
			{
				return stringValue;
			}
		}
	}
}
