using System;
using System.Collections.Generic;
using System.Reflection;
using System.Text;

using NMock2.Internal;

using Castle.Core.Interceptor;
using Castle.DynamicProxy;

namespace NMock2.Monitoring
{
	internal class MockObjectInterceptor : MockObject, IInterceptor
	{
		private static readonly Dictionary<MethodInfo, object> mockObjectMethods = new Dictionary<MethodInfo, object>();
		private MockStyle mockStyle;

		static MockObjectInterceptor()
		{
			// We want to be able to quickly recognize any later invocations
			// on methods that belong to IMockObject or IInvokable, so we cache
			// their definitions here.
			foreach(MethodInfo methodInfo in typeof(IMockObject).GetMethods())
			{
				mockObjectMethods.Add(methodInfo, null);
			}

			foreach(MethodInfo methodInfo in typeof(IInvokable).GetMethods())
			{
				mockObjectMethods.Add(methodInfo, null);
			}
		}



		public MockObjectInterceptor(	Mockery mockery,
										Type mockedType,
										string name,
										MockStyle mockStyle) : base(mockery, mockedType, name)
		{
			this.mockStyle = mockStyle;
		}



		#region IInterceptor Members

        public void Intercept(IInvocation interceptedInvocation)
        {
            // Check for calls to basic NMock2 infrastructure
            if (mockObjectMethods.ContainsKey(interceptedInvocation.Method))
            {
                try
                {
                    interceptedInvocation.ReturnValue = interceptedInvocation.Method.Invoke(this, interceptedInvocation.Arguments);
                }
                catch (TargetInvocationException tie)
                {
                    // replace stack trace with original stack trace
                    FieldInfo remoteStackTraceString = typeof(Exception).GetField("_remoteStackTraceString", BindingFlags.Instance | BindingFlags.NonPublic);
                    remoteStackTraceString.SetValue(tie.InnerException, tie.InnerException.StackTrace + Environment.NewLine);
                    throw tie.InnerException;
                }

                return;
            }

			// Ok, this call is targeting a member of the mocked class
			object invocationTarget = MockedType.IsInterface ? interceptedInvocation.Proxy : interceptedInvocation.InvocationTarget;
			Invocation invocationForMock = new Invocation(	invocationTarget,
															interceptedInvocation.Method,
															interceptedInvocation.Arguments);

			if(ShouldCallInvokeImplementation(invocationForMock))
			{
				interceptedInvocation.Proceed();
				return;
			}

			interceptedInvocation.ReturnValue = ProcessCallAgainstExpectations(invocationForMock);
		}

		#endregion

		private bool ShouldCallInvokeImplementation(Invocation invocationForMock)
		{
			return !MockedType.IsInterface
				&& mockStyle == MockStyle.Transparent
				&& !Mockery.HasExpectationFor(invocationForMock);
		}

		private object ProcessCallAgainstExpectations(Invocation invocationForMock)
		{
			Invoke(invocationForMock);

			if(invocationForMock.IsThrowing)
				throw invocationForMock.Exception;

			if(invocationForMock.Result == Missing.Value && invocationForMock.Method.ReturnType != typeof(void))
			{
				throw new InvalidOperationException(
					string.Format("You have to set the return value for method '{0}' on '{1}' mock.",
								  invocationForMock.Method.Name, invocationForMock.Method.DeclaringType.Name));
			}

			return invocationForMock.Result;
		}
	}
}
